-- Huffman Encoding Program:
--         written by Junko C. Miura at RDS
--         Public Domain
--
--         It compresses an input file using Huffman Coding Scheme and creates
--         an encoded compressed file.
--
-- Usage:     ex huff inFileName outFileName
--         or
--            ex huff inFileName               <-- create default compressed 
--                                                 file HUFF.OUT
--         or
--            ex huff                          <-- to enter the file names 
--
-- where inFileName:  file name that you want to compress
--       outFileName: file name that you create by compressing inFileName
--
-- How it Works:
--    1. It reads the entire input file to determine the frequency of each 
--       8-bit character.
--    2. It re-assigns the bit-codes for each character, in order of frequency.
--       The most frequent characters are assigned very short codes, maybe 
--       1, 2 or 3 bits, while the least frequent characters are assigned 
--       longer codes, maybe 8, 9 or more bits.
--    3. It outputs a tree (Euphoria sequence) to the header of the output 
--       file. This tree shows which bit pattern is assigned to which character.
--    4. It re-reads the entire input file and generates the body of the 
--       output file, substituting the new codes.
-- 
-- Bug Fix: March 17, 1999 - now outputs an extra blank in the file after the 
--                         - tree so it would work with Euphoria 2.1 get()
--

without type_check
include machine.e
include file.e
include sort.e
include huffman.e

constant FREQ_INDX = 1, LETT_INDX = 2, LEFT_CHILD = 2,
	 RIGHT_CHILD = 3
constant LEAF_NODE = 2 

sequence huffman, bitPattern
atom totalNumBits
sequence command

procedure openFiles()
-- get file name to be encoded and open it 

    command = command_line()
    if length(command) = 3 then         -- if user typed only one file name
	command = append(command, "HUFF.OUT") -- assume default output file
    elsif length(command) < 3 then
	command = append(command, 
			 get_response("file name to be compressed? "))
	command = append(command,
			 get_response("result file name? (hit ENTER for " &
				      "the default (ie, HUFF.OUT)) "))  
	if length(command[4]) = 0 then
	    command[4] = "HUFF.OUT"
	end if
    end if
	
    plainFile = open(command[3], "rb")
    if plainFile = -1 then
	errorPut("couldn't open " & command[3])
    end if
    encodeFile = open(command[4], "wb")
    if encodeFile = -1 then
	errorPut("couldn't open " & command[4])
    end if
end procedure

procedure buildFreq()
    integer c, j
    
    -- get statistics (frequency of characters) from the input file
    huffman = repeat({0, 0}, NUM_ASCII)
    for i = 1 to NUM_ASCII do
	huffman[i][LETT_INDX] = i - 1
    end for
    
    openFiles()
    while TRUE do
	c = getc(plainFile)
	if c != EOF then
	    c += 1
	    huffman[c][FREQ_INDX] += 1
	else
	    exit
	end if
    end while
    
    -- sort the array and discard the first entries with 0 frequencies. The 
    -- result will be the initial Huffman trees each with only a leaf node.
    huffman = sort(huffman)
    j = 1
    while j <= NUM_ASCII do
	if huffman[j][FREQ_INDX] != 0 then
	    exit
	end if
	j += 1
    end while
    if j > NUM_ASCII then
	-- empty input file, so just get out. Empty output file was created.
	abort(0)
    end if
    huffman = huffman[j..NUM_ASCII]
end procedure

procedure huffInsert(atom freq, integer leftChild, integer rightChild)
    integer j
    
    j = rightChild + 1        -- parent has larger frequency value than a child
    while j <= length(huffman) do
	if huffman[j][FREQ_INDX] >= freq then
	    exit
	end if
	j += 1
    end while
    huffman = append(huffman[1..(j - 1)], {freq, leftChild, rightChild}) & 
	      huffman[j..length(huffman)]        -- ok even when j = length()+1
end procedure

procedure buildHuffTree()
    -- pick the next least weighed two nodes, create a parent node, and insert
    -- it into an appropriate place in the huffman array. When finished going
    -- through with only one node left, that is the root node of the completed
    -- Huffman tree.
    integer j
    sequence temp1, temp2
    
    j = 1
    while j < length(huffman) do
	temp1 = huffman[j]
	temp2 = huffman[j + 1]

	huffInsert(temp1[FREQ_INDX] + temp2[FREQ_INDX], j, j + 1)
	j += 2
    end while
    -- at this moment, huffman[length] contains the root of completed tree
end procedure

function traverse(sequence root, integer depth, sequence bitPtrn)
    -- given a root node, traverse a tree and accumulate the total path weight. 
    -- Also compute encoding bit pattern for each ASCII character. This is a 
    -- recursive function.
    
    if length(root) = LEAF_NODE then     -- root is a leaf node
	totalNumBits = totalNumBits + root[FREQ_INDX] * depth
	bitPattern[root[LETT_INDX] + 1] = bitPtrn
	return root[LETT_INDX]
    else                                 -- root is an internal node
	return {
		traverse(huffman[root[LEFT_CHILD]], depth + 1, 
			 bitPtrn & GO_LEFT_CHILD)
	       ,
		traverse(huffman[root[RIGHT_CHILD]], depth + 1, 
			 bitPtrn & GO_RIGHT_CHILD)
	       }
    end if
end function

procedure getHuffDetail()
    -- compute the total number of bits needed in the encoded data. Also 
    -- compute encoding bit pattern for each ASCII character. Index to 
    -- bitPattern is (ASCII code + 1), while its elememt value is atom 0 for 
    -- an ASCII character of frequency 0, or a sequence of bit patterns for 
    -- an ASCII character with frequency > 0.
    
    bitPattern = repeat(0, NUM_ASCII)
    
    if length(huffman) = 1 then     
	-- this is the special case of only 1 leaf root node in the Huffman 
	-- tree. It's when only 1 character was used in the input file. Note
	-- that the same character for left and right leaf nodes can be used.
	totalNumBits = huffman[1][FREQ_INDX] 
	bitPattern[huffman[1][LETT_INDX] + 1] = { GO_LEFT_CHILD }
	tree = { huffman[1][LETT_INDX], huffman[1][LETT_INDX] }
	
    else   -- normal case
	-- start with the root node (depth is 0) of the completed Huffman tree.
	totalNumBits = 0
	tree  = traverse(huffman[length(huffman)], 0, {})     
    end if
end procedure

procedure putHuffData()
    integer c, num
    sequence bits
    
    -- output file header, that is: the completed Huffman tree and the number
    -- of significant bits in the last byte of the encoded file. The last byte
    -- will be padded with bits 0's that are to be ignored.
    c = remainder(totalNumBits, BYTE_SIZE)
    if c = 0 then
	c = BYTE_SIZE
    end if
    puts(encodeFile, c)
    print(encodeFile, tree)
    puts(encodeFile, ' ') -- leave blank after top-level object
    
    if seek(plainFile, 0) != 0 then
	errorPut("seek() failed")
    end if

    -- output encoded body
    bits = {}
    while TRUE do
	c = getc(plainFile)
	if c != EOF then
	    bits = bits & bitPattern[c + 1]
	    while length(bits) > BYTE_SIZE do
		num = bits_to_int(bits[1..BYTE_SIZE])
		puts(encodeFile, num)
		bits = bits[(BYTE_SIZE + 1)..length(bits)]
	    end while
	else
	    bits = bits & repeat(0, BYTE_SIZE - length(bits))
	    num = bits_to_int(bits)
	    puts(encodeFile, num)
	    exit
	end if
    end while
    
end procedure

procedure statistics()
    sequence d
    integer oldsize, newsize
    
    d = dir(command[3])
    oldsize = d[1][D_SIZE]
    d = dir(command[4])
    newsize = d[1][D_SIZE]
    printf(DISPLAY, "reduced to %.2f %% of original size\n", 
	   100 * newsize / oldsize)
end procedure

-- start of main body
buildFreq()
buildHuffTree()
getHuffDetail()
putHuffData()
closeFiles()
statistics()

